package ladasha.processes;

import java.io.*;
import javax.sound.midi.*;
import java.util.*;
import ladasha.statistics.*;
import ladasha.io.*;

public class Humanizer {
  // singleton
  private static Humanizer instance = null;
  
  private Humanizer() {}
  
  public static Humanizer getInstance() {
    logger.log("obtaining instance of Humanizer");
    if (instance == null) instance = new Humanizer();
    return instance;
  }
  
  public static void setMidiFile(File newFile) {
    logger.log("Humanizer: setMidiFile(newFile): setting MIDI file");
    if (newFile == null) return;
    midiFile = newFile;
  } 
  
  public static void setProfiles(HashMap<String,File> newProfiles) {
    logger.log("Humanizer: setProfiles(newProfiles): setting profile list");
    if (newProfiles == null) return;
    profiles = newProfiles;
    // debug
    /*
    for(String s : profiles.keySet()) {
      System.out.println(s + " -> " + profiles.get(s).toString());
    }
    */
  }
  
  // fields
  private static File midiFile = null;
  private static HashMap<String,File> profiles = null;
  private Sequence sequenceOrig = null;
  private Sequence sequenceNew = null;
  private Sequencer seq1 = null;
  private static DiscreteRNG discreteRNG = DiscreteRNG.getInstance();
  private static GaussianRNG gaussianRNG = GaussianRNG.getInstance();
  private static ProfileHandler profileHandler = ProfileHandler.getInstance();
  private static Double[] gaussian = new Double[2];
  private static Logger logger = Logger.getInstance();

  private static final int VELO_MASK = 0b1001;
  private static final int AMP_MASK = 0b1010;
  private static final int TIME_MASK = VELO_MASK; // same events
  
  private static volatile int passcount = 0;
  private static volatile int passescompleted = 0;

  private static volatile boolean initialized = false;

  public static int getPasscount() {
    return passcount;
  }
  
  public static int getPassesCompleted() {
    return passescompleted;
  }

  public static HashMap<String,File> getProfiles() {
    return profiles;
  }

  public static File getMidiFile() {
    return midiFile;
  }

	public Sequence getMidiSequence() {
		if (sequenceNew == null) return null;
		else return sequenceNew;
	}

	public Sequencer getMidiSequencer() {
		if (seq1 == null) return null;
		else return seq1;
	}
  // methods
  public int go() {
    passcount = 0;
		new Thread(new Runnable() {
      public void run() {
        logger.log("Humanizer: go(): going");
        modifySequence();
      }
    }).start();
    for(;;) {
      //System.out.println("initialized: " + initialized);
      if(initialized) break;
    }
    return getPasscount();
  }
  
  public void previewStop() {
    logger.log("Humanizer: previewStop(): stopping preview");
    if(seq1 == null) return; // throw exception here in the future
    try {
      seq1.stop();
    } catch (Exception ex) { ex.printStackTrace(); }
  }
  
  public void previewStart() {
    logger.log("Humanizer: previewStart(): starting preview");
    try{
      seq1 = MidiSystem.getSequencer();
      if(seq1 == null) {
        return; // throw exception here in the future
      } else { 
        // get resources to get midi
        seq1.open();
      }
      // sequencer gets file
      seq1.setSequence(sequenceNew);
      Thread t = new Thread(new Runnable() {
      	public void run() { 
      	  seq1.start();  
        }
      });
      t.start();
    } catch(Exception ex){
      ex.printStackTrace();
    }
  }

  public boolean isPlaying() {
    return seq1.isRunning();
  }

  public void saveFile(String filename) {
    logger.log("Humanizer: saveFile(filename): saving MIDI file");
    if (sequenceNew == null) return; // throw exception here in the future
    try {
      MidiSystem.write(sequenceNew, 1, new File(filename));
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
  
  private void requestSequence() {
    logger.log("Humanizer: requestSequence(): requesting MIDI sequence");
    MidiFileHandler mhandle = MidiFileHandler.getInstance();
    if (!mhandle.loadMidiFile(this.midiFile))
      // throw exception here in the future
      return;
    sequenceOrig = mhandle.getMidiSequence();
  }
  
  public boolean humanizationCompleted() {
    return passescompleted == passcount;
  }
  
  private void modifySequence() {
    for(int i = 0; i < 5; i++) { // try to humanize up to 5 times
      logger.log("Humanizer: modifySequence(): modifying MIDI sequence");
      initialized = false;
      requestSequence();
      initialized = true;
      sequenceNew = sequenceOrig; // copy the original sequence
      passcount = profiles.size();
      passescompleted = 0;
      for (String s : profiles.keySet()) {
        modifySequence(s, profiles.get(s));
        passescompleted++;
      }
      if (validateResults()) break;
    }
  }
  
  
  private void modifySequence(String selection, File profile) {
    // string in the form of: ##X (where ## is a channel number, and X is a parameter)
    
    String attr = selection.substring(selection.length()-1).toUpperCase();
    int trackNum = Integer.parseInt(selection.substring(0,selection.length()-1))-1;
    
    boolean isGaussian = false;
    
    if (profile == null) return;
    ProfileHandler.setProfile(profile);
    if (ProfileHandler.isDiscrete()) {
      HashMap<Double,Double> discrete = ProfileHandler.getDiscrete();
      DiscreteRNG.setProbabilities(discrete);
    } else if (ProfileHandler.isGaussian()) {
      gaussian = ProfileHandler.getGaussian();
      isGaussian = true;
    } else return;
    
    
    // requires Java 7 (faster, but uglier)
    switch (attr) {
      case "A":
        modifyAmplitude(trackNum, profile, isGaussian);
        break;
      case "T":
        modifyTiming(trackNum, profile, isGaussian);
        break;
      case "V":
        modifyVelocity(trackNum, profile, isGaussian);
        break;
      default:
        // throw exception here in the future
        return;
    }
    
    // non-Java 7 version (slower, cleaner, portable)
    /*
    if (attr.equalsIgnoreCase("A"))
      modifyAmplitude(trackNum, profile);
    else if (attr.equalsIgnoreCase("T"))
      modifyTiming(trackNum,profile);
    else if (attr.equalsIgnoreCase("V"))
      modifyVelocity(trackNum, profile);
    else
      // throw exception here in the future
      return;
    */
  }
  
  // MIDI Messages of interest:
  // XXXX is channel #
  // KKKKKKK is note #
  // VVVVVVV is velocity
  // PPPPPPP is key pressure
  // 1001XXXX 0KKKKKKK 0VVVVVVV -> Note On (for Velocity)
  // 1010XXXX 0KKKKKKK 0PPPPPPP -> Polyphonic Key Pressure (for Amplitude)
  // for timing, manipulate MidiEvent.setTick() directly
  
  private void modifyVelocity(int tracknum, File profile, boolean isGaussian) {
    logger.log("Humanizer: modifyVelocity(tracknum,profile,isGaussian): modifying channel " + tracknum + " with profile " + profile.getName());    
    Track[] oldTracks = sequenceNew.getTracks();
    Track[] newTracks = new Track[oldTracks.length];
    for (int i = 0; i < oldTracks.length; i++) {
      Track trackold = oldTracks[i];
      sequenceNew.deleteTrack(trackold);
      newTracks[i] = sequenceNew.createTrack();
      int numevents = trackold.size();
      for (int ii = 0; ii < numevents; ii++) {
        MidiEvent event = trackold.get(ii);
        MidiMessage message = event.getMessage();
        int statusbyte = message.getStatus() >>> 4;
        //System.out.println("statusbyte: " + statusbyte + " velomask: " + VELO_MASK);
        if (statusbyte != VELO_MASK) { // not a velocity command
          newTracks[i].add(event);
        }
        else { // a velocity command
          if((statusbyte & 0b1111) == tracknum) {
            logger.log("   -> velocity event!");
            byte[] data = message.getMessage();
            if (isGaussian) {
              Double[] p = ProfileHandler.getGaussian();    
              data[2] += (int)gaussianRNG.next(p[0],p[1]);
            } else {
              data[2] += (int)discreteRNG.next();
            }
            data[2] &= ~(0b10000000); // constrain values
            long time = event.getTick();
            //System.out.println("time: " + time);
            //System.out.println("data: " + data[0] + " " + data[1] + " " + data[2]);            
            
            try {
              event = new MidiEvent(new ShortMessage(data[0],data[1],data[2]),time);
            } catch (Exception ex) {ex.printStackTrace();}
            
          }
          newTracks[i].add(event);
        }
      }
    }
  }
  
  private void modifyAmplitude(int tracknum, File profile, boolean isGaussian) {  
    logger.log("Humanizer: modifyAmplitude(tracknum,profile,isGaussian): modifying channel " + tracknum + " with profile " + profile.getName());
    Track[] oldTracks = sequenceNew.getTracks();
    Track[] newTracks = new Track[oldTracks.length];
    for (int i = 0; i < oldTracks.length; i++) {
      Track trackold = oldTracks[i];
      sequenceNew.deleteTrack(trackold);
      newTracks[i] = sequenceNew.createTrack();
      int numevents = trackold.size();
      for (int ii = 0; ii < numevents; ii++) {
        MidiEvent event = trackold.get(ii);
        MidiMessage message = event.getMessage();
        int statusbyte = message.getStatus() >>> 4;
        //System.out.println("statusbyte: " + statusbyte + " ampmask: " + AMP_MASK);
        if (statusbyte != AMP_MASK) { // not an amplitude command
          newTracks[i].add(event);
        }
        else { // an amplitude command
          if((statusbyte & 0b1111) == tracknum) {
            logger.log("   -> amplitude event!");
            byte[] data = message.getMessage();
            if (isGaussian) {
              Double[] p = ProfileHandler.getGaussian();
              data[2] += (int)gaussianRNG.next(p[0],p[1]);
            } else {
              data[2] += (int)discreteRNG.next();
            }
            data[2] &= ~(0b10000000); // constrain values
            long time = event.getTick();
            //System.out.println("time: " + time);
            //System.out.println("data: " + data[0] + " " + data[1] + " " + data[2]);            
            
            try {
              event = new MidiEvent(new ShortMessage(data[0],data[1],data[2]),time);
            } catch (Exception ex) {ex.printStackTrace();}
            
          }
          newTracks[i].add(event);
        }
      }
    }
  }
  
  private void modifyTiming(int tracknum, File profile, boolean isGaussian) {
    logger.log("Humanizer: modifyTiming(tracknum,profile,isGaussian): modifying channel " + tracknum + " with profile " + profile.getName());
    Track[] oldTracks = sequenceNew.getTracks();
    Track[] newTracks = new Track[oldTracks.length];
    for (int i = 0; i < oldTracks.length; i++) {
      Track trackold = oldTracks[i];
      sequenceNew.deleteTrack(trackold);
      newTracks[i] = sequenceNew.createTrack();
      int numevents = trackold.size();
      for (int ii = 0; ii < numevents; ii++) {
        MidiEvent event = trackold.get(ii);
        MidiMessage message = event.getMessage();
        int statusbyte = message.getStatus() >>> 4;
        //System.out.println("statusbyte: " + statusbyte + " timemask: " + TIME_MASK);
        if (statusbyte != TIME_MASK) { // not a timing command
          newTracks[i].add(event);
        }
        else { // a timing command
          if((statusbyte & 0b1111) == tracknum) {
            logger.log("   -> timing event!");
            long time = event.getTick();
            if (isGaussian) {
	      Double[] p = ProfileHandler.getGaussian();
              time += (long)gaussianRNG.next(p[0],p[1]);
            } else {
              time += (long)discreteRNG.next();
            }
            //System.out.println("time: " + time);
            //System.out.println("data: " + data[0] + " " + data[1] + " " + data[2]);            
            
            try {
              event = new MidiEvent(message,time);
            } catch (Exception ex) {ex.printStackTrace();}
            
          }
          newTracks[i].add(event);
        }
      }
    }
  }
  
  private boolean validateResults() {
    logger.log("Humanizer: validateResults(): validating results...");
    long threshold = 5000000; // +/- 5 seconds
    boolean retval = false;
    if (Math.abs(sequenceOrig.getMicrosecondLength() -
                 sequenceNew.getMicrosecondLength()) <= threshold)
      retval = true;
    logger.log("   -> " + retval);   
    return retval;
  }
  
}
